/*
 * Copyright (c) 2011 LAAS/CNRS                       --  Oct 2011
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 * Description: low-pass, Kalman and other Bayesian filters for perception
 *  
 *  
 *  
 *  Creation date : Jan 19 22:35:29 2011 
 *  Author: HE Wuwei
 *  LAAS-CNRS, Toulouse FRANCE
 */







#include<iostream>
#include<string>
#include<fstream>
#include<sstream>
#include<vector>

#include "lowPassFilter.hpp"

using namespace std;

namespace sensFusion_filters
{


LPfilter::LPfilter()
{
  
  this->tInit_ = 0;
  this->timePeriod_ = 0;
  this->historySize_ = 50; //historySize by default
  this->firstActWindow_ = 0;
  this->ticInit_ = 0;
  this->ticTimeNow_ = 0;
  
  for(unsigned i=0; i != this->historySize_; ++i){
    this->rawHistory_.push_back(0.0);
    this->filHistory_.push_back(0.0);
  }
}

LPfilter::LPfilter(const LPfilter& cf)
 {
  
  this->tInit_ = cf.tInit_;
  this->timePeriod_ = cf.timePeriod_;
  this->historySize_ = cf.historySize_;
  this->firstActWindow_ = cf.firstActWindow_;
  this->ticInit_ = cf.ticInit_;
  this->ticTimeNow_ = cf.ticTimeNow_;


  this->rawHistory_.resize(historySize_);
  this->filHistory_.resize(historySize_);

  for(unsigned i=0; i != this->historySize_; ++i){
    this->rawHistory_.at(i) = cf.rawHistory_.at(i);
    this->filHistory_.at(i) = cf.filHistory_.at(i);
  }
}

void LPfilter::init(double tInit, double timePeriod, unsigned int historySize, unsigned int firstActWindow)
{
  tInit_ = tInit;
  timePeriod_ = timePeriod;
  historySize_ = historySize;
  firstActWindow_ = firstActWindow;
  ticInit_ = 0;
  ticTimeNow_ = 0;

  rawHistory_.resize(historySize_);
  filHistory_.resize(historySize_);
}

void LPfilter::setTime(double& tInit) 
{
  tInit_ = tInit;
  ticInit_ = 0;
  ticTimeNow_ = 0;
}

void LPfilter::pushBackSample(double x, double t)
{
  double newPos = x;

  rawHistory_.push_back(newPos);
  rawHistory_.pop_front();
}


void LPfilter::doFiltering()
{
							
  ticTimeNow_ = ticTimeNow_ + 1;
 
  if( (ticTimeNow_ - ticInit_) > firstActWindow_) {

    double newVal = 0.0;

    //a simple 4th order low-pass filter
    newVal = 1 * (  0.026080129725961530978706193195648666006  * rawHistory_.at(historySize_ - 1)
		    +0.1825197334643963997802273979687015526    * rawHistory_.at(historySize_ - 2)
		    +0.413160497616284150268484154366888105869  * rawHistory_.at(historySize_ - 3)
		    +0.413160497616284150268484154366888105869  * rawHistory_.at(historySize_ - 4)
		    +0.1825197334643963997802273979687015526    * rawHistory_.at(historySize_ - 5)
		    +0.026080129725961530978706193195648666006  * rawHistory_.at(historySize_ - 6)
		    );

    filHistory_.push_back(newVal);
    filHistory_.pop_front();
  }
  
  else {
    filHistory_.push_back(rawHistory_.at(historySize_-1));
    filHistory_.pop_front();
  }
}
  

void LPfilter::getFilteredResults(double& x)
{

  if( (ticTimeNow_ - ticInit_) > firstActWindow_) {
    x = filHistory_.at(historySize_ - 1) / 1.245; //static gain
  }
  else {
    x = filHistory_.at(historySize_ - 1); 
  }
}






/******************************************************************
                3D Low Pass Filter
 *************************************************************** */



LPfilter3D::LPfilter3D()
{
  this->vecSize_ = 4;
  this->tInit_ = 0;
  this->timePeriod_ = 0;
  this->historySize_ = 50; //historySize by default
  this->firstActWindow_ = 0;
  this->ticInit_ = 0;
  this->ticTimeNow_ = 0;
  
  for(unsigned i=0; i != this->historySize_; ++i){
    this->rawHistory_.push_back(std::vector<double>(4, 0.0));
    this->filHistory_.push_back(std::vector<double>(4, 0.0));
  }
}

LPfilter3D::LPfilter3D(const LPfilter3D& cf)
 {
  this->vecSize_ = cf.vecSize_;
  this->tInit_ = cf.tInit_;
  this->timePeriod_ = cf.timePeriod_;
  this->historySize_ = cf.historySize_;
  this->firstActWindow_ = cf.firstActWindow_;
  this->ticInit_ = cf.ticInit_;
  this->ticTimeNow_ = cf.ticTimeNow_;


  this->rawHistory_.resize(historySize_);
  this->filHistory_.resize(historySize_);
  for(unsigned i=0; i != this->historySize_; ++i){
    this->rawHistory_.at(i).resize(vecSize_);
    this->filHistory_.at(i).resize(vecSize_);
    for(unsigned j=0; j != this->vecSize_; ++j){
      this->rawHistory_.at(i).at(j) = cf.rawHistory_.at(i).at(j);
      this->filHistory_.at(i).at(j) = cf.filHistory_.at(i).at(j);
    }
  }
}

void LPfilter3D::init(double tInit, double timePeriod, unsigned int historySize, unsigned int firstActWindow)
{
  tInit_ = tInit;
  timePeriod_ = timePeriod;
  historySize_ = historySize;
  firstActWindow_ = firstActWindow;
  ticInit_ = 0;
  ticTimeNow_ = 0;

  rawHistory_.resize(historySize_);
  filHistory_.resize(historySize_);
  
}

void LPfilter3D::setTime(double& tInit) 
{
  tInit_ = tInit;
  ticInit_ = 0;
  ticTimeNow_ = 0;
}

void LPfilter3D::pushBackSample(double x, double y, double z, double t)
{
  vector<double> newPos(4);
  newPos.at(0) = x;
  newPos.at(1) = y;
  newPos.at(2) = z;
  newPos.at(3) = t;

  rawHistory_.push_back(newPos);
  rawHistory_.pop_front();
}


void LPfilter3D::doFiltering(double t)
{
							
  //  std::cout << "start filtering ... " << std::endl;

  ticTimeNow_ = ticTimeNow_ + 1;

  if( (ticTimeNow_ - ticInit_) > firstActWindow_) {
    
    vector<double> newFilPos(4);

 newFilPos.at(0)= 1 * (
		       0.026080129725961530978706193195648666006  * rawHistory_.at(historySize_-1).at(0)
		       +0.1825197334643963997802273979687015526    * rawHistory_.at(historySize_-2).at(0)
		       +0.413160497616284150268484154366888105869  * rawHistory_.at(historySize_-3).at(0)
		       +0.413160497616284150268484154366888105869  * rawHistory_.at(historySize_-4).at(0)
		       +0.1825197334643963997802273979687015526    * rawHistory_.at(historySize_-5).at(0)
		       +0.026080129725961530978706193195648666006  * rawHistory_.at(historySize_-6).at(0)
		       );

 newFilPos.at(1)= 1 * (
		       0.026080129725961530978706193195648666006  * rawHistory_.at(historySize_-1).at(1)
		       +0.1825197334643963997802273979687015526    * rawHistory_.at(historySize_-2).at(1)
		       +0.413160497616284150268484154366888105869  * rawHistory_.at(historySize_-3).at(1)
		       +0.413160497616284150268484154366888105869  * rawHistory_.at(historySize_-4).at(1)
		       +0.1825197334643963997802273979687015526    * rawHistory_.at(historySize_-5).at(1)
		       +0.026080129725961530978706193195648666006  * rawHistory_.at(historySize_-6).at(1)
		       );

 newFilPos.at(2)= 1 * (
		       0.026080129725961530978706193195648666006  * rawHistory_.at(historySize_-1).at(2)
		       +0.1825197334643963997802273979687015526    * rawHistory_.at(historySize_-2).at(2)
		       +0.413160497616284150268484154366888105869  * rawHistory_.at(historySize_-3).at(2)
		       +0.413160497616284150268484154366888105869  * rawHistory_.at(historySize_-4).at(2)
		       +0.1825197334643963997802273979687015526    * rawHistory_.at(historySize_-5).at(2)
		       +0.026080129725961530978706193195648666006  * rawHistory_.at(historySize_-6).at(2)
		       );

    filHistory_.push_back(newFilPos);
    filHistory_.pop_front();
    
    
  }
  
  else {

    filHistory_.push_back(rawHistory_.at(historySize_-1));
    filHistory_.pop_front();
  }
}
  

void LPfilter3D::getFilteredResults(double& x, double& y, double& z, double& t)
{

  if( (ticTimeNow_ - ticInit_) > firstActWindow_) {
    x = filHistory_.at(historySize_-1).at(0)  / 1.245; //static gain
    y = filHistory_.at(historySize_-1).at(1)  / 1.245;
    z = filHistory_.at(historySize_-1).at(2)  / 1.245;
    t = filHistory_.at(historySize_-1).at(3);
  }
  else {


    x = filHistory_.at(historySize_-1).at(0); 
    y = filHistory_.at(historySize_-1).at(1); 
    z = filHistory_.at(historySize_-1).at(2);
    t = filHistory_.at(historySize_-1).at(3);
  }
}


} //namespace
